//
// Copyright (C) 2000-2002 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source$
// $Revision: 2187 $
// $Date: 2006-12-15 03:17:56 +0300 (Пт, 15 дек 2006) $
// Author: Andrey Slepuhin <pooh@msu.ru>

#ifndef __pxx_chunk_allocator_hh__
#define __pxx_chunk_allocator_hh__

#include "pxx_allocator.hh"
#include "pxx_default_allocator.hh"
#include "pxx_chunk_allocator_block_header.hh"

namespace pxx
{

extern size_t chunks_in_block ;

template <typename type_t>
class ChunkAllocator
{

private:

  struct FreeList
  {
    //
    // Pointer to the previous item
    FreeList* prev ;
    //
    // Pointer to the next item
    FreeList* next ;
    //
    //
    // Remove item from the list
    inline void remove () ;

  };
  //
  // A reference to underlying memory allocator
  Allocator& allocator ;
  //
  // List of memory blocks
  CABlockHeader block_list ;
  //
  // List of ofree chunks
  FreeList free_list ;
  //
  // Chunk size
  size_t chunk_size ;

  size_t block_size ;

  size_t nchunks ;

  CABlockHeader* root ;

public:
  //
  // Constructor
  inline ChunkAllocator (Allocator& _allocator = default_allocator) ;
  //
  // Constructor with explicit number of chunks in block
  inline ChunkAllocator (
    size_t _nchunks,
    Allocator& _allocator = default_allocator
  ) ;
  //
  // Destructor
  inline ~ChunkAllocator () ;
  //
  // Allocate one chunk
  inline type_t* allocate () ;

  inline void deallocate (type_t* _ptr) ;

private:

  inline void init () ;
  inline CABlockHeader* get_block (void* _ptr) ;
  void init_block (CABlockHeader* _ptr) ;
  void free_block (CABlockHeader* _ptr) ;

};

}

#endif // __pxx_chunk_allocator_hh__
