//
// Copyright (C) 2000-2002 Andrey Slepuhin <pooh@msu.ru>
//
// libp++ is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// libp++ is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libp++; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// $Source$
// $Revision: 2187 $
// $Date: 2006-12-15 03:17:56 +0300 (Пт, 15 дек 2006) $
// Author: Andrey Slepuhin <pooh@msu.ru>

#ifndef __pxx_heap_hh__
#define __pxx_heap_hh__

#include "pxx_common.hh"
#include "pxx_sys_error.hh"

namespace pxx
{

//
// A class representing generic mmapped block of memory of size 2^n.
class Heap
{

protected:
  //
  // Heap start address, if zero then address will be chosen by mmap()
  void* start_addr ;
  //
  // Maximum heap size, if zero then heap may grow without limits
  size_t max_size ;
  //
  // Binary logarithm of maximum size
  size_t max_size_log ;
  //
  // Current heap size
  size_t current_size ;
  //
  // Binary logarithm of current size
  size_t cur_size_log ;
  //
  // File descriptor for mmapped file, if it is equal -1, then use MAP_ANON
  int fd ;

  NO_COPY_CTOR(Heap)
  NO_ASSIGN(Heap)

public:
  //
  // Constructor
  Heap (
    size_t _init_size = 0,      // Initial heap size, page size by default
    size_t const _max_size = 0, // Maximum size, not limited by default
    void* const _start = null,
    int _fd = -1
  ) throw (SysError) ;
  //
  // Destructor
  ~Heap () throw (SysError) ;
  //
  // Expand a heap
  void expand (unsigned _order = 1) throw (SysError) ;
  //
  // Get heap start address
  inline void* get_start_addr () const ;

};

}

#endif // __pxx_heap_hh__

