// $Id: rf_macros.hh 3897 2008-09-05 12:56:35Z orlov $

#ifndef __rf_macros_hh__
#define __rf_macros_hh__

#include "rf_closure.hh"
#include "rf_retval.hh"

namespace rfrt
{

extern Expr* __rf_call_saved_top;

extern rftype::Closure* __rf_closure_to_call;

#ifndef RFRT_STINGY_STACK

extern bool __keep_stack;

#define RF_start_push   stack->start_push_mode()
#define RF_stop_push    stack->stop_push_mode()
#define RF_start_pop    stack->start_pop_mode()
#define RF_stop_pop     stack->stop_pop_mode()
#define RF_start_args
#define RF_start_ress   stack->start_ress_mode()
#define RF_stop_ress(f) stack->stop_ress_mode(f)
#define RF_stack_up     stack->up()

#else // RFRT_STINGY_STACK

#define RF_start_push   Expr* __rf_call_saved_top = stack->get_top()
#define RF_stop_push    stack->set_top(__rf_call_saved_top)
#define RF_start_pop    stack->start_pop_mode(__rf_call_saved_top)
#define RF_stop_pop     stack->set_top(__rf_call_saved_top); stack->dec_depth()
#define RF_start_args   Expr* __saved_top = stack->get_top()
#define RF_start_ress   stack->set_top(__saved_top)
#define RF_stop_ress(f) stack->inc_depth()
#define RF_stack_up     stack->set_top(__saved_top); stack->dec_depth()

#endif // RFRT_STINGY_STACK


#ifndef RFRT_USE_STATEMENT_EXPR

#ifdef RFRT_STINGY_STACK
#undef  RF_start_push
#define RF_start_push   __rf_call_saved_top = stack->get_top()
#endif

#ifndef RFRT_DISABLE_TAIL_RECURSION

static inline bool __rf_call (rftype::Func const* _f)
{
#ifdef RFRT_STINGY_STACK
  Expr* _top1 = __rf_call_saved_top;
#endif
  while (true) {
    RetVal __retval = (*_f)();
    if (!__retval.is_tail()) {
#ifdef RFRT_STINGY_STACK
      __rf_call_saved_top = _top1;
#endif
      return __retval;
    } else {
      _f = __retval.get_ptr();
    }
  }
}

#else // RFRT_DISABLE_TAIL_RECURSION

static inline bool __rf_call (rftype::Func const* _f)
{
#ifdef RFRT_STINGY_STACK
  Expr* _top1 = __rf_call_saved_top;
#endif
  RetVal __retval = (*_f)();
#ifdef RFRT_STINGY_STACK
  __rf_call_saved_top = _top1;
#endif
  return __retval;
}

#endif // RFRT_DISABLE_TAIL_RECURSION

#define RF_CALL(f,args,ress) (\
  RF_start_push, \
  stack->push(args), \
  RF_stop_push, \
  __rf_call(f) ? \
    RF_start_pop, \
    stack->pop(ress), \
    RF_stop_pop, \
    true : false \
)

#else // RFRT_USE_STATEMENT_EXPR

#ifndef RFRT_DISABLE_TAIL_RECURSION

#define RF_CALL(f,args,ress) ({\
  RF_start_push; \
  stack->push(args); \
  RF_stop_push; \
  RetVal __retval; \
  rftype::Func* __p = f; \
  while (true) { \
    __retval = (*__p)(); \
    if (!__retval.is_tail()) break; \
    __p = __retval.get_ptr(); \
  } \
  if (__retval) {\
    RF_start_pop; \
    stack->pop(ress); \
    RF_stop_pop; \
  } \
  (bool)__retval; \
})

#else // RFRT_DISABLE_TAIL_RECURSION

#define RF_CALL(f,args,ress) ({\
  RF_start_push; \
  stack->push(args); \
  RF_stop_push; \
  RetVal __retval = (*f)(); \
  if (__retval) {\
    RF_start_pop; \
    stack->pop(ress); \
    RF_stop_pop; \
  } \
  (bool)__retval; \
})

#endif // RFRT_DISABLE_TAIL_RECURSION

#endif // RFRT_USE_STATEMENT_EXPR


#ifndef RFRT_DISABLE_TAIL_RECURSION

#define RF_TAILCALL(f,args,ress) \
  RF_stack_up; \
  RF_start_push; \
  stack->push(args); \
  RF_stop_push; \
  return f;

#else // RFRT_DISABLE_TAIL_RECURSION

#define RF_TAILCALL(f,args,ress) \
  RF_stack_up; \
  RF_start_push; \
  stack->push(args); \
  RF_stop_push; \
  return (*f)();

#endif // RFRT_DISABLE_TAIL_RECURSION


#define RF_APPLY(f,args,ress) \
  rfrt::__rf_closure_to_call = f; \
  RF_CALL(f->get_func(), args, ress);

#define RF_TAILAPPLY(f,args,ress) \
  rfrt::__rf_closure_to_call = f; \
  RF_TAILCALL(f->get_func(), args, ress);


#define RF_FUNC(f,args,ress) \
  namespace rffunc { RetVal f##_func_ (void); } \
  rftype::Func* f = new rftype::Func(rffunc::f##_func_, WSTRPREFIX#f); \
  RetVal rffunc::f##_func_ (void) { \
    RF_start_args; \
    for args { \
      RF_start_ress; \
      for ress { \
        RF_stop_ress(f);

#define RF_END \
        return true; \
      } \
    } \
  }

#define RF_CLOSURE(f,params,args,ress) \
  RF_FUNC(f, args, ress) \
  for params { \
    rfrt::__rf_closure_to_call->pop_arg();

#define RF_END_CLOSURE \
    RF_END \
  }

#if 0
#define RF_FUNC(f,args,ress) \
  namespace rffunc { \
    RetVal f##_func_ (void); \
    wchar_t const f##_name_[] = L"\0" WSTRPREFIX#f; \
  } \
  rftype::StaticObject<rftype::Func>* f = \
      new rftype::StaticObject<rftype::Func>( \
            rffunc::f##_func_, rffunc::f##_name_, wcslen(WSTRPREFIX#f) + 1); \
  RetVal rffunc::f##_func_ (void) { \
    RF_start_args; \
    for args { \
      RF_start_ress; \
      for ress { \
        RF_stop_ress(f);

#define RF_END \
        return true; \
      } \
    } \
  }
#endif

#if 0
#define RF_DECL(f) \
  extern RetVal f (void)
#endif
#define RF_DECL(f) \
  extern rftype::Func* f;

#define RF_ARG rfrt::Arg
#define RF_RES rfrt::Result

#define RF_VOID (;;)

#define RF_ARG0 RF_VOID
#define RF_ARG1(a) (RF_ARG a;;)
#define RF_ARG2(a,b) (RF_ARG a,b;;)
#define RF_ARG3(a,b,c) (RF_ARG a,b,c;;)
#define RF_ARG4(a,b,c,d) (RF_ARG a,b,c,d;;)
#define RF_ARG5(a,b,c,d,e) (RF_ARG a,b,c,d,e;;)
#define RF_ARG6(a,b,c,d,e,f) (RF_ARG a,b,c,d,e,f;;)
#define RF_ARG7(a,b,c,d,e,f,g) (RF_ARG a,b,c,d,e,f,g;;)
#define RF_ARG8(a,b,c,d,e,f,g,h) (RF_ARG a,b,c,d,e,f,g,h;;)
#define RF_ARG9(a,b,c,d,e,f,g,h,i) (RF_ARG a,b,c,d,e,f,g,h,i;;)

#define RF_RES0 RF_VOID
#define RF_RES1(a) (RF_RES a;;)
#define RF_RES2(a,b) (RF_RES a,b;;)
#define RF_RES3(a,b,c) (RF_RES a,b,c;;)
#define RF_RES4(a,b,c,d) (RF_RES a,b,c,d;;)
#define RF_RES5(a,b,c,d,e) (RF_RES a,b,c,d,e;;)
#define RF_RES6(a,b,c,d,e,f) (RF_RES a,b,c,d,e,f;;)
#define RF_RES7(a,b,c,d,e,f,g) (RF_RES a,b,c,d,e,f,g;;)
#define RF_RES8(a,b,c,d,e,f,g,h) (RF_RES a,b,c,d,e,f,g,h;;)
#define RF_RES9(a,b,c,d,e,f,g,h,i) (RF_RES a,b,c,d,e,f,g,h,i;;)

#ifndef RFRT_STINGY_STACK
# define RF_RETFAIL \
    return false
#else
# define RF_RETFAIL \
    return __saved_top
#endif

#ifndef RFRT_STINGY_STACK
# define RF_TRAP \
    size_t __trap_depth   = stack->get_depth(); try
#else
# define RF_TRAP \
    Expr*  __trap_top   = stack->get_top(); \
    size_t __trap_depth = stack->get_depth(); try
#endif

#define RF_WITH catch (Expr const& err)

#ifndef RFRT_STINGY_STACK
# define RF_CLEANUP \
    do { \
      __keep_stack = false; \
      stack->cleanup(__trap_depth); \
    } while (0)
# ifdef RFRT_COMPATIBLE
#  define RF_ERROR(x) \
     do { \
       if (trap_stack_on) __keep_stack = true; \
       Expr y = x; \
       throw Expr(y, y.get_len() - 2, 2); \
     } while (0)
# else
#  define RF_ERROR(x) \
     do { \
       if (trap_stack_on) __keep_stack = true; \
       throw (x); \
     } while (0)
# endif
#else
# define RF_CLEANUP stack->destroy_results(__trap_top, __trap_depth)
# define RF_ERROR(x) throw (x)
#endif

#define RF_LIB_ERROR(msg) RF_FUNC_ERROR(Expr::create<Word>(msg))

#ifndef RFRT_STINGY_STACK
# define RF_FUNC_ERROR(msg) \
    do { \
      if (trap_stack_on) __keep_stack = true; \
      wchar_t const* fname; \
      size_t len = stack->get_func(stack->get_depth())->get_name(&fname); \
      throw Expr::create<Word>(fname, len) + msg; \
    } while (0)
#else
# define RF_FUNC_ERROR(msg) \
    throw Expr::create<Word>(__FUNCTION__) + msg
#endif

#define RF_LIB_ERROR_IN_FUNC(func, msg) \
  throw Word::create_expr(func) + Word::create_expr(msg)

#if 0
#define RF_LIB_ERROR_IN_FUNC(func, msg) \
  { \
    Expr _res; \
    RF_CALL(rfp__mangle::Cpp2Rfp, \
           (rftype::Word::Word(func).to_symbol()), \
           (_res)); \
    RF_CALL(Convert::To__Chars, (_res), (_res)); \
    RF_CALL(Convert::To__Upper, (_res), (_res)); \
    RF_CALL(Convert::To__Word,  (_res), (_res)); \
    RF_ERROR(_res + rftype::Word::Word(msg).to_symbol()); \
  }
#endif

#define RF_NEW_SYMBOL(CppType, TermClass) \
  template<> \
  struct SymbolType< CppType > \
  { \
    /*enum { type = c };*/ \
    typedef TermClass Type; \
  }

#define RF_SYMBOL(CppType) Symbol<CppType, SymbolType< CppType >::Type>

#define RF_REGISTER_SYMBOL(CppType) \
  template <> \
  const TypeRegister RF_SYMBOL(CppType)::reg(&RF_SYMBOL(CppType)::funcs)

#define RF_REGISTER_SYMBOL_VARIANT(CppType, InstCppType) \
  template <> \
  const TypeRegister RF_SYMBOL(CppType)::reg( \
      RF_SYMBOL(InstCppType)::get_type(), \
      &RF_SYMBOL(CppType)::funcs \
  )

}

#endif // __rf_macros_hh__
