// $Source$
// $Revision: 2187 $
// $Date: 2006-12-15 03:17:56 +0300 (Пт, 15 дек 2006) $
// $Author: orlov $

#ifndef __rf_symbol_hh__
#define __rf_symbol_hh__

#include "rf_term.hh"
#include "rf_types.ih"  // for TypeRegister()

namespace rftype
{

using namespace rfrt ;

template<size_t _size>
class CheckShortObjectSize ;

template<>
class CheckShortObjectSize<sizeof(intptr_t)>
{};

#if UINTPTR_MAX != UINT32_MAX
  template<>
  class CheckShortObjectSize<4>
  {};  
#endif

template<>
class CheckShortObjectSize<2>
{};

template<>
class CheckShortObjectSize<1>
{};


//template<class C, term_class_t tc = object_term>
//class Symbol;
template<class C, class Type>
class Symbol;

template<class C>
class Symbol<C, FlatTerm> :
  CheckShortObjectSize<sizeof(C)>,
  public Term
{

public:

  inline Symbol () ;

  template<typename CppType1>
  inline Symbol (CppType1 const& _arg1) ;

  template<typename CppType1, typename CppType2>
  inline Symbol (CppType1 const& _arg1, CppType2 const& _arg2) ;

  template<typename CppType1, typename CppType2, typename CppType3>
  inline Symbol (
      CppType1 const& _arg1, CppType2 const& _arg2, CppType3 const& _arg3) ;

  inline C const& get_object () const ;

  static void ctor (Symbol* _to, Symbol const* _from) ;
  static void dtor (Symbol* _term) ;
  static bool eq (Symbol const* _term1, Symbol const* _term2) ;
  static int compare (Symbol const* _term1, Symbol const* _term2) ;
  static uint32_t hash (Symbol const* _term) ;
  static pxx::WString to_string (Symbol const* _term) ;
  static size_t get_name (Symbol const* _s, wchar_t const** _np) ;

  static short_type_funcs_t funcs ;
  static const TypeRegister reg ;

  static inline unsigned get_type () ;

  template<typename CppType>
  static inline size_t compute_length (CppType const* _array) ;
};

template<class C>
class Symbol<C, StaticTerm> :
  public Term
{

  NO_COPY_CTOR(Symbol)
  NO_ASSIGN(Symbol)

public:

  inline Symbol (C* _p) ;

  inline C* get_obj_ptr () const ;

  static void ctor (Symbol* _to, Symbol const* _from) ;
  static void dtor (Symbol* _term) ;
  static bool eq (Symbol const* _term1, Symbol const* _term2) ;
  static int compare (Symbol const* _term1, Symbol const* _term2) ;
  static uint32_t hash (Symbol const* _term) ;
  static pxx::WString to_string (Symbol const* _term) ;
  static size_t get_name (Symbol const* _s, wchar_t const** _np) ;

  static short_type_funcs_t funcs ;

  static const TypeRegister reg ;

  static inline unsigned get_type() ;

};

template<class C>
class Symbol<C, ObjectTerm> :
  public Term
{

  struct Wrapper
  {
    size_t ref_count;
    C object;
  };

  NO_COPY_CTOR(Symbol)
  NO_ASSIGN(Symbol)

public:

  inline Symbol () ;

  template<typename CppType1>
  inline Symbol (CppType1 const& _arg1) ;

  template<typename CppType1, typename CppType2>
  inline Symbol (CppType1 const& _arg1, CppType2 const& _arg2) ;

  template<typename CppType1, typename CppType2, typename CppType3>
  inline Symbol (
      CppType1 const& _arg1, CppType2 const& _arg2, CppType3 const& _arg3) ;

  inline void replace_object () ;

  template<typename CppType1>
  inline void replace_object (CppType1 const& _arg1) ;

  template<typename CppType1, typename CppType2>
  inline void replace_object (CppType1 const& _arg1, CppType2 const& _arg2) ;

  template<typename CppType1, typename CppType2, typename CppType3>
  inline void replace_object (
      CppType1 const& _arg1, CppType2 const& _arg2, CppType3 const& _arg3) ;

  inline C const& get_object () const ;

  inline C* get_obj_ptr () const ;

  static void ctor (Symbol* _to, Symbol const* _from) ;
  static void dtor (Symbol* _term) ;
  static bool eq (Symbol const* _term1, Symbol const* _term2) ;
  static int compare (Symbol const* _term1, Symbol const* _term2) ;
  static uint32_t hash (Symbol const* _term) ;
  static pxx::WString to_string (Symbol const* _term) ;
  static size_t get_name (Symbol const* _s, wchar_t const** _np) ;

  static short_type_funcs_t funcs ;

  static const TypeRegister reg ;

  static inline unsigned get_type() ;

  /// FIXME: transitional code!
  //  After moving all term types to Symbol<> term_class_t should be completely
  //  removed.
  //
  /// Term class for template instantiations
  static const term_class_t term_class = short_term;
};

template<class C>
struct SymbolType;

template<class C>
class NamedObject :
  public C
{
public:
  pxx::WString name;
  
  inline NamedObject (wchar_t const* _name) : C(), name(_name) {};
  
  template<typename CppType1>
  inline NamedObject (wchar_t const* _name, CppType1 _arg1) :
    C(_arg1), name(_name)
  {};
};

template<class C>
class Symbol<NamedObject<C>, ObjectTerm> :
  public Term
{

  struct Wrapper
  {
    size_t ref_count;
    NamedObject<C> object;
  };

  NO_COPY_CTOR(Symbol)
  NO_ASSIGN(Symbol)

public:

  inline Symbol () ;

  template<typename CppType1>
  inline Symbol (CppType1 _arg1) ;

  template<typename CppType1, typename CppType2>
  inline Symbol (CppType1 _arg1, CppType2 _arg2) ;

  inline NamedObject<C> const& get_object () const ;

  inline NamedObject<C>* get_obj_ptr () const ;

  static void ctor (Symbol* _to, Symbol const* _from) ;
  static void dtor (Symbol* _term) ;
  static bool eq (Symbol const* _term1, Symbol const* _term2) ;
  static int compare (Symbol const* _term1, Symbol const* _term2) ;
  static uint32_t hash (Symbol const* _term) ;
  static pxx::WString to_string (Symbol const* _term) ;
  static size_t get_name (Symbol const* _s, wchar_t const** _np) ;

  static short_type_funcs_t funcs ;

  static const TypeRegister reg ;

  static inline unsigned get_type() ;

};

}

#endif // __rf_symbol_hh__

